package com.bergerkiller.bukkit.nolagg.datapool;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

public class Pool <T> {
	private static List<Pool<?>> pools = new ArrayList<Pool<?>>(2);
	private static boolean cleanupRunning = false;
	private static Thread thread;
	public static void init() {
		cleanupRunning = true;
		thread = new Thread() {
			public void run() {
				while (!this.isInterrupted() && cleanupRunning) {
					try {
						synchronized (pools) {
							for (Pool<?> pool : pools) {
								pool.cleanUp();
							}
						}
						Thread.sleep(10000);
					} catch (InterruptedException ex) {}
				}
			}
		};
		thread.start();
	}
	public static void deinit() {
		cleanupRunning = false;
		thread.interrupt();
		synchronized (pools) {
			for (Pool<?> pool : pools) {
				pool.clear();
			}
			pools.clear();
		}
	}
	
	public Pool(final int stepSize) {
		synchronized (pools) {
			pools.add(this);
		}
		this.stepSize = stepSize;
	}
	
	private Queue<PoolBuffer<T>> freeBuffers = new LinkedList<PoolBuffer<T>>();
	private ArrayList<PoolBuffer<T>> buffers = new ArrayList<PoolBuffer<T>>();
	private boolean increased = false;
	private final int stepSize;
	public void cleanUp() {
		synchronized (this) {
			if (this.increased) {
				this.increased = false;
				return;
			}
			Iterator<PoolBuffer<T>> iter = this.buffers.iterator();
			PoolBuffer<T> buff;
			int i = 0;
			while (iter.hasNext()) {
				buff = iter.next();
				if (buff.isFree()) {
					iter.remove();
				} else {
					buff.index = i++;
				}
			}
			freeBuffers.clear();
		}
	}
	
	protected void setFree(PoolBuffer<T> buffer) {
		synchronized (this) {
			this.freeBuffers.offer(buffer);
		}
	}

	public void clear() {
		synchronized (this) {
			for (PoolBuffer<T> buff : this.buffers) {
				buff.setFree();
			}
			this.buffers.clear();
			this.freeBuffers.clear();
		}
	}
	public int size() {
		synchronized (this) {
			return this.buffers.size();
		}
	}
	
	public PoolBuffer<T> create(PoolData<T> reference) {
		PoolBuffer<T> buff;
		synchronized (this) {
			//ask for a recently freed buffer
			buff = this.freeBuffers.poll();
			if (buff != null) {
				if (buff.tryLoad(reference)) {					
					return buff;
				}
			}
			//re-use a non-freed buffer if possible
			Iterator<PoolBuffer<T>> iter = this.buffers.iterator();
			while (iter.hasNext()) {
				buff = iter.next();
				if (buff.tryLoad(reference)) {
					return buff;
				}
			}
		}
		//generate a new batch of buffers
		//load a sane amount of buffers
		this.increased = true;
		buff = new PoolBuffer<T>(this, reference.getNewData(), reference);
		synchronized (this) {
			buff.index = this.buffers.size();
			this.buffers.add(buff);
			for (int i = 0; i < this.stepSize - 1; i++) {
				PoolBuffer<T> tmpbuff = new PoolBuffer<T>(this, reference.getNewData(), reference);
				tmpbuff.setFree();
				tmpbuff.index = this.buffers.size();
				this.buffers.add(tmpbuff);
			}
		}
		return buff;
	}

}
