package com.bergerkiller.bukkit.nolagg.scheduler;

import java.util.Set;
import java.util.logging.Level;

import com.bergerkiller.bukkit.nolagg.NoLagg;
import com.bergerkiller.bukkit.nolagg.sending.ChunkSendQueue;

import net.minecraft.server.Block;
import net.minecraft.server.Chunk;
import net.minecraft.server.WorldServer;

@SuppressWarnings("rawtypes")
public class ChunkOperation implements Comparable {
	private final Type mode;
	public final Chunk c;
	public Set ticks;
	public ChunkOperation(final Chunk chunk, final Type type) {
		this.c = chunk;
		this.mode = type;
	}
	
	public static enum Type{LIGHTING};
	
	public int compareTo(Object object) {
		if (object instanceof ChunkOperation) {
			return ((ChunkOperation) object).mode.ordinal() - this.mode.ordinal();
		} else {
			return 0;
		}
	}
	
	public Type getMode() {
		return this.mode;
	}
		
	/*
	 * Whatever needs to happen when this object is being scheduled (synced)
	 */
	public void preexecute() {
		if (this.mode == Type.LIGHTING) {
			//remove queued chunks for this chunk
		    ChunkSendQueue.deQueueGlobal(this.c);
		}
	}
		
	public WorldServer getWorld() {
		return (WorldServer) this.c.world;
	}
		
	private static int getLightLevel(final Chunk chunk, int x, final int y, int z) {
		if (y <= 0 || y >= 127) return 0;
		if (x >= 0 && z >= 0 && x < 16 && z < 16) {
			return chunk.h.a(x, y, z);
		}
		x += chunk.x << 4;
		z += chunk.z << 4;
		if (((WorldServer) chunk.world).chunkProviderServer.isChunkLoaded(x >> 4, z >> 4)) {
			return chunk.world.getLightLevel(x, y, z);
		}
		return 0;
	}
	public static boolean fixBlockLighting(final Chunk chunk, final int x, final int y, final int z) {
		final int typeid = chunk.getTypeId(x, y, z);
		if (!Block.o[typeid]) {
			final int factor = Math.max(1, Block.q[typeid]);
			final int light = chunk.h.a(x, y, z);
			//actual editing here
			int newlight = light + factor;
			newlight = Math.max(newlight, getLightLevel(chunk, x - 1, y, z));
			newlight = Math.max(newlight, getLightLevel(chunk, x + 1, y, z));
			newlight = Math.max(newlight, getLightLevel(chunk, x, y, z - 1));
			newlight = Math.max(newlight, getLightLevel(chunk, x, y, z + 1));
			newlight = Math.max(newlight, getLightLevel(chunk, x, y - 1, z));
			newlight = Math.max(newlight, getLightLevel(chunk, x, y + 1, z));
			newlight -= factor;
			//when exceeding in loops, just pick the highest value
			if (newlight > light) {
				chunk.h.a(x, y, z, newlight);
				return true;
			}
		}
		return false;
	}
	
	/*
	 * What needs to happen when the scheduler finds this object
	 */
	public void execute() {
		//1 = fix lighting
		//2 = save - unload mode
		//3 = save - autosave mode
		//save
		if (this.mode == Type.LIGHTING) {
			c.h();
			c.initLighting();
			boolean haserror = true;
			int loops = -1;
			int x, y, z;
			while (haserror) {
				if (loops > 20) {
					NoLagg.log(Level.WARNING, "Failed to fix all lighting issues in chunk [" + c.x + "/" + c.z + "/" + c.world.getWorld().getName() + "]");
					break;
				}
				haserror = false;
				loops++;
				for (x = 0; x < 16; x++) {
					for (z = 0; z < 16; z++) {
						for (y = c.b(x,  z); y > 0; --y) {
							if (fixBlockLighting(c, x, y, z)) {
								haserror = true;
							}
						}
					}
				}
			}
			//queue
			ChunkSendQueue.scheduleQueueGlobal(this.c);
		}
	}
	
}
